function [bits_out] = RLF_decode(bits_in, block, blocks, error_map, key)
%
% bits_out = RLF_decode(bits_in, block, blocks, error_map, key)
%
% Implementation of the random linear fountain decoder. Uses Gaussian
% elimination for solving the problem.
% 
% Input Parameters:
%
%   - bits_in - input bitstream, vector of 0 and 1
%
%   - block   - symbol length
%
%   - blocks  - the number of original message symbols
%  
%   - error_map - the tampering map, vector of 0 and 1, tampered as 1-s
%
%   - key - integer, the seed for the PRNG
%
% Output:
% 
% bits_out - decoded bitstream, vector of 0 and 1
%
% ------------------------------------------------------------------------
%
% Copyright (C) 2011 Paweł Korus <pkorus@agh.edu.pl>
%
% Version: 2011.12.16
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation version 3.0.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%

    blk = length(bits_in)/block;
    D = zeros(1, blk);
    G = zeros(blk, blocks);
    
    rand('twister', key);
    
    for d = 1:blk
        ro = randint(1, blocks, 2); 
        G(d,:) = ro;
        if nargin > 3
            if error_map(d) == true
                G(d,:) = 0;
            end
        end
    end
    
    if rank(G) < blocks
        throw( MException('rlc:insufficientData', 'Not enought input symbols'));
    end

    % Append symbols to G
    D = zeros(length(bits_in)/block, blocks+block);
    for bi=1:length(bits_in)/block
        D(bi,:) = [G(bi,:) bits_in(1+(bi-1)*block:bi*block)];
    end
    
    G = D;
    
    % Step I
    for i = 1:blocks
        % find rows with ones on i-th position
        rows = i-1+find(G(i:end,i));
        if length(rows) == 0
            throw( MException('rlc:insufficientData', 'Not enought input symbols'));
        end
        % select the one with the least ones
        sums = sum(G');
        relevant_sums = sums(rows);
        [x, row_ind] = min(relevant_sums);
        row_ind = rows(row_ind);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
        % move the row as the first one
        G(row_ind,:) = G(i,:);
        G(i,:) = tr;
    end
    
    % Step II - back-substitution
        for i = 1:blocks
        % find rows with ones on i-th position
        rows = find(G(1:blocks-i+1,blocks-i+1));
        row_ind = rows(end);
        tr = G(row_ind,:);
        % remove the dependency on row_ind form other rows
        for b = rows'
            if b ~= row_ind
                G(b,:) = mod(G(b,:) + tr,2);
            end
        end
    end
    
    
    bits_out = G(1:blocks,blocks+1:end);
    bits_out = bits_out';
    bits_out = bits_out(:)';

end