function [img_low_quality, psnr] = get_reference(img, q_bits)
%
% [img_low_quality, psnr] = get_reference(img, q_bits)
%
% Gets the reference image for a specified number of reference bits per block
%
% Input parameters:
%
%   - img - uint8 matrix representing a single channel grayscale image with
%           the size of multiples of 8
%
%   - q_bits - number of bits of rec. reference per image block (between 1
%
% Output:
% 
%   - img_low_quality - the resulting reference image
%   - psnr - the peak signal to noise ratio
% ------------------------------------------------------------------------
%
% Copyright (C) 2011 Paweł Korus <pkorus@agh.edu.pl>
%
% Version: 2011.12.16
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation version 3.0.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%

    [H, W] = size(img);
            
    % PREPARE A LOW-DETAIL VERSION OF THE IMAGE.
    dr_factor = 8;
    block = 8;

    % Limit the dynamic range of the image
    img_low_dr = double(floor(double(img)/dr_factor));

    f = @(x) dct2(x);
    fi = @(x) idct2(x);

    S_b = blkproc(img_low_dr, block*[1, 1], f);

    % Encode the reconstruction stream

    % pointer to the reconstruction stream data
    bi = 1;

    % Matrix P contains the number of bits that can be used to
    % represent each coefficient in a block, for maximum quality
    P = get_reference_alloc(q_bits);

    for bx = 1:W/block
        for by = 1:H/block
            B = S_b(1+(by-1)*block:(by)*block, 1+(bx-1)*block:(bx)*block);
            B = round(B);            

            % Matrix X is the maxmum absolute value that can be encoded for
            % each coefficient, as defined by P
            M = pow2(P-1)-1;
            M(1,1) = pow2(P(1,1))-1;

            % Saturate the values that exceed M
            B(B > M) = M(B > M);
            B(B < -M) = -(M(B < -M));

            % Zero the coefficients that have no anticipated bits
            B(P == 0) = 0;

            S_b(1+(by-1)*block:(by)*block, 1+(bx-1)*block:(bx)*block) = B;
        end
    end

    img_low_quality = dr_factor*uint8(blkproc(S_b, block*[1, 1], fi));
    psnr = calc_psnr(img, img_low_quality);
end