function r = reconstruction_test(img, q_bits, t_rate, method, tampering_pattern, tampering_source)
%
% r = reconstruction_test(img, q_bits, t_rate, method, tampering_pattern,
%                         tampering_source)
%
% Implementation of the reference image authentication algorithm based on
% the self-recovery model of content reconstruction. It executed a full
% protection-tampering-authentication-reconstruction process.
%
% Input parameters:
%
%   - img - uint8 matrix representing a single channel grayscale image with
%           the size of multiples of 8
%
%   - q_bits - number of bits of rec. reference per image block (between 1
%              and 160)
%
%   - t_rate - tampering rate, float [0,1]
%
%   - method - string,defines the desired payload decoding method:
%
%     'self-recovery'   - the proposed self-recovery model
%     'general-erasure' - traditional random linear fountain decoding with
%                         Gaussian elimination
%
%   - tampering_pattern - string, defines the desired tampering pattern
%
%     'continuous' - modify T first image blocks
%     'random'     - random tampering pattern
%
%   - tampering_source - string, defines the source for the tampering
%                        optional, defaults to 'random'
%
%     'random'    - adds random noise to the image
%     <file-path> - loads a custom image from disk, will be adapted to
%                   match the dimensions of the cover image
%
% Output:
% 
% r - structure with the following fields:
%                 ref_psnr: PSNR of the reconstruction reference, zeroed
%                           LSB's
%                     psnr: PSNR of the watermarked image
%                    gamma: the final tampering rate
%                 rec_psnr: PSNR of the reconstructed image, with
%                           randomized LSB's in restored areas
%    reconstruction_demand: the reconstruction demand
%                        B: number of blocks in the image
%                        K: number of reconstruction symbols
%                        T: number of tampered image blocks
%                        E: errors in the tampering map
%               str_errors: errors in the decoded reference stream
%                error_map: vector with the error map, f. of block index
%                     imgt: the tampered image
%                    imgt2: the tampered image with tinted tamp. detection
%                  img_fin: the final reconstructed image
%                  img_ref: the reconstruction reference
% 
% ------------------------------------------------------------------------
%
% Copyright (C) 2011 Paweł Korus <pkorus@agh.edu.pl>
%
% Version: 2011.12.16
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation version 3.0.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%

    [W, H] = size(img);
    r = struct();
    
    dr_factor = 8;
    BLOCK = 8;
    S_LENGTH = 160;
    RANDOMIZE_LSBS = false;
    KEY = 100*sum(clock());
            
    % PREPARE A LOW-DETAIL VERSION OF THE IMAGE.

    % Limit the dynamic range of the image
    img_low_dr = double(floor(double(img)/dr_factor));

    % Transform the low-dynamic range image into spectral domain
    f = @(x) dct2(x);
    fi = @(x) idct2(x);

    S_b = blkproc(img_low_dr, BLOCK*[1, 1], f);

    % Get a mapping for the zig-zag order
    indices = zigzag(BLOCK);

    % bits needed for the reconstruction stream
    bitstream = zeros(1,(H/BLOCK*W/BLOCK)*q_bits);

    % GENERATE THE RECONSTRUCTION REFERENCE

    % Matrix P contains the number of bits that can be used to
    % represent each coefficient in a block, for maximum quality
    P = get_reference_alloc(q_bits);

    % pointer to the reconstruction stream data
    bi = 1;

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK
            B = S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            B = round(B);

            % Matrix M is the maxmum absolute value that can be encoded for
            % each coefficient, as defined by P
            M = pow2(P-1)-1;
            M(1,1) = pow2(P(1,1))-1;

            % Saturate the values that exceed M
            B(B > M) = M(B > M);
            B(B < -M) = -(M(B < -M));

            % Zero the coefficients that have no anticipated bits
            B(P == 0) = 0;

            % Encode the coefficients into a bitstream
            V = B(indices);
            Q = P(indices);
            Q(Q == 0) = [];
            bits = false(1, sum(sum(P)));
            last_bits_index = 0;

            for i = 1:length(Q)
                v = V(i);
                q = Q(i);
                if i == 1
                    bits(last_bits_index+1:last_bits_index+q) = dec2binvec(v, q);
                elseif i > 1
                    bits(last_bits_index+1:last_bits_index+q) = [v<0 dec2binvec(abs(v), q - 1)];
                end
                last_bits_index = last_bits_index + q;    
            end

            bitstream(bi:bi+length(bits)-1) = bits;
            bi = bi+length(bits);

            S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
        end
    end

    % Get a spatial-domain representation of the reference image
    img_low_quality = dr_factor*uint8(blkproc(S_b, BLOCK*[1, 1], fi));
    
    % Calculate reconstruction reference quality
    r.ref_psnr = calc_psnr(img, img_low_quality);
    
    % ENCODE THE RECONSTRUCTION STREAM WITH THE RLF CODE
    capacity = 160*W*H/BLOCK/BLOCK;
    symbols_needed = capacity/S_LENGTH;
    bitstream_enc = RLF_encode(bitstream, S_LENGTH, symbols_needed, KEY);

    % EMBED THE WATERMARK

    imgw = img;

    blocks_no = ceil(length(bitstream)/S_LENGTH);
    fprintf('Input symbols: %d\n', blocks_no);
    fprintf('Output symbols: %d\n', symbols_needed);
    fprintf('Payload / Capacity Ratio: %.3f\n', blocks_no/symbols_needed);

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK
            % Determine block ID in scale 1, ..., WH/b/b
            block_id = by + (bx-1)*H/BLOCK - 1 + 1;

            % Get block data for original and low-rate versions
            B = img(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            B_a = uint8(img_low_dr(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK));

            % Take:
            % 1) block ID,
            % 2) low quality block data (for reconstruction),
            % 3) low dynamic range block data (for tampering detection).
            % ...
            bits_b = dec2binvec(block_id);
            bits_d = bitstream_enc(1+(block_id-1)*160:block_id*160);
            bits_a = zeros(1, 5*numel(B_a));
            for i=1:numel(B_a)
                bits_a(1+(i-1)*5:i*5) = dec2binvec(B_a(i), 5);
            end

            % and supply it into a hashing function
            h = md5([bits_b bits_a bits_d]);

            % Since classic MD5 yields 128 bit output, make a short vesion (32
            % bit) by summming up binary symbol data
            h_short = zeros(1, 32);
            for i=1:length(h)
                h_short(i) = mod(uint8(h(i)), 2);
            end

            % Replace 3 LSBs with the hash value and low quality block data
            bits = [h_short bits_d];

            for i=1:numel(B_a)
                B(i) = 8*B_a(i) + binvec2dec(bits(1+(i-1)*3:i*3));
            end

            imgw(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
        end
    end

    % Calculate the watermarked image PSNR
    r.psnr = calc_psnr(img, imgw);

    % TAMPER IMAGE BLOCKS
    imgt = imgw;    
    rand('twister', sum(100*clock));
    
    % Determine the number of tampered blocks from the tampering rate
    T = int32(W*H/BLOCK/BLOCK*t_rate);
    r.gamma = double(T) / (W*H/BLOCK/BLOCK);
    
    % Configure the tampering source as either an external image or the 
    % original one with transposed blocks and random noise
    
    if ~exist('tampering_source', 'var')
        tampering_source = 'random';
    end
    
    if strcmp(tampering_source, 'random')
        imgs = imresize(imgw', size(imgw));
        imgs = imgs + uint8(randint(size(img,1),size(img,2),32));
    else            
        if ~exist(tampering_source, 'file')
            fprintf('ERROR File does not exist!\n');
            return;                
        end
            
        imgs = imread(tampering_source);
        
        if size(imgs,3) > 1
            imgs = uint8(mean(imgs,3));
        end

        imgs = imresize(imgs, size(imgw));
    end
    
    % Generate the desired tampering pattern
    if strcmp(tampering_pattern, 'random')
        
        fprintf('INFO Using random tampering pattern with rate %.2f\n', r.gamma);
        tampered_blocks = randperm(W*H/BLOCK/BLOCK);
        tampered_blocks = tampered_blocks(1:T);
        
    elseif strcmp(tampering_pattern, 'continuous')
        
        fprintf('INFO Using continuous tampering pattern with rate %.2f\n', r.gamma);
        tampered_blocks = 1:T;
    else
        fprintf('WARNING Unknown tampering pattern\n');
        return;
    end
    
    % Perform the actual tampering
    for bx = 1:W/BLOCK
        
        for by = 1:H/BLOCK
            
            block_id = by + (bx-1)*H/BLOCK - 1 + 1;
            
            if numel(find(tampered_blocks == block_id)) > 0
                B = imgs(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
                imgt(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
            end
            
        end
        
    end

    % DETECT TAMPERED REGIONS

    imgt2 = imgt;

    bits_recovered = false(1, W/BLOCK*H/BLOCK*160);
    error_map = false(1, W/BLOCK*H/BLOCK);

    for bx = 1:W/BLOCK
        for by = 1:H/BLOCK

            % Generate block id in range 1, ..., WH/b/b
            block_id = by + (bx-1)*H/BLOCK - 1 + 1;

            % Get block data from the image
            B = imgt(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
            % Generate low dynamic range version of the data
            B_a = floor(double(B)/dr_factor);

            % Read the embedded bits from the image
            bits = zeros(1, 3*BLOCK*BLOCK);
            for i=1:numel(B_a)
                V = dec2binvec(B(i), 8);
                bits(1+(i-1)*3:i*3) = V(1:3);
            end

            % Read the embedded hash value
            h_short_read = bits(1:32);

            % Regenerate the bit sequence used to feed the hashing function
            bits_d = bits(33:end);
            bits_b = dec2binvec(block_id);
            bits_a = zeros(1, 5*numel(B_a));
            for i=1:numel(B_a)
                bits_a(1+(i-1)*5:i*5) = dec2binvec(B_a(i), 5);
            end

            % Perform hashing
            h = md5([bits_b bits_a bits_d]);
            h_short = zeros(1, 32);
            for i=1:length(h)
                h_short(i) = mod(uint8(h(i)), 2);
            end

            % If hashes do not match, the block has been tampered with.
            % In such case:
            % - mark the block as erased (model of channel with erasures)
            % - set the bits recovered from this block to 0
            if ~all(h_short == h_short_read)
                imgt2(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = 1.5*B;
                bits_recovered(1+(block_id-1)*160:block_id*160) = 0;
                error_map(block_id) = true;
            else
                bits_recovered(1+(block_id-1)*160:block_id*160) = bits_d;
            end
        end
    end

    fprintf('Tampered blocks: %d (%.2f %%)\n', sum(sum(error_map)), 100*sum(sum(error_map))/(W/BLOCK*H/BLOCK));

    % Count errors in error map detection
    error_map_temp = error_map;
    error_map_temp(tampered_blocks) = 0;
    E = sum(error_map_temp);
    
    % Pre-initialize helper variables
    r.rec_psnr = 0;
    ERR = 0;
    img_fin = uint8(zeros(size(img)));
    
    if E == 0
        
        % OBTAIN THE RECONSTRUCTION REFERENCES OF AUTHENTIC BLOCKS
        % & THE RECONSTRUCTION MAP - WHICH INPUT SYMBOLS ARE NEEDED
        reconstruction_map = zeros(1, blocks_no);

        % pointer to the reconstruction stream data
        bi = 1;
        bitstream_ref = zeros(1,(H/BLOCK*W/BLOCK)*q_bits);

        % Limit the dynamic range of the image
        img_low_dr = double(floor(double(imgt)/dr_factor));
        S_b = blkproc(img_low_dr, BLOCK*[1, 1], f);
        
        % Matrix P contains the number of bits that can be used to
        % represent each coefficient in a block, for maximum quality
        P = get_reference_alloc(q_bits);
    
        for bx = 1:W/BLOCK
            for by = 1:H/BLOCK
                block_id = by + (bx-1)*H/BLOCK - 1 + 1;
                B = S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
                B = round(B);

                % Matrix M is the maxmum absolute value that can be encoded for
                % each coefficient, as defined by P
                M = pow2(P-1)-1;
                M(1,1) = pow2(P(1,1))-1;

                % Saturate the values that exceed M
                B(B > M) = M(B > M);
                B(B < -M) = -(M(B < -M));

                % Zero the coefficients that have no anticipated bits
                B(P == 0) = 0;

                % Encode the coefficients into a bitstream
                V = B(indices);
                Q = P(indices);
                Q(Q == 0) = [];
                bits = false(1, sum(sum(P)));
                last_bits_index = 0;

                if error_map(block_id) == false
                    for i = 1:length(Q)
                        v = V(i);
                        q = Q(i);
                        if i == 1
                            bits(last_bits_index+1:last_bits_index+q) = dec2binvec(v, q);
                        elseif i > 1
                            bits(last_bits_index+1:last_bits_index+q) = [v<0 dec2binvec(abs(v), q - 1)];
                        end
                        last_bits_index = last_bits_index + q;    
                    end
                else
                    f_block_id = 1+floor((bi-1)/ S_LENGTH);
                    l_block_id = 1+floor((bi+length(bits)-2)/S_LENGTH);
                    for m_block_id = f_block_id:l_block_id
                        reconstruction_map(m_block_id) = 1;
                    end
                end

                bitstream_ref(bi:bi+length(bits)-1) = bits;
                bi = bi+length(bits);

                S_b(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
            end
        end

        try
            r.reconstruction_demand = mean(reconstruction_map);
            fprintf('Reconstruction demand: %.2f\n', mean(reconstruction_map));            
            
            % Decode the necessary reference symbols
            if strcmp(method, 'general-erasure')
                fprintf('Using general erasure model for reference stream decoding.\n');
                bits_decoder = RLF_decode(bits_recovered, S_LENGTH, blocks_no, error_map, KEY);
            elseif strcmp(method, 'self-recovery')
                fprintf('Using self-recovery model for reference stream decoding.\n');
                bits_decoder = RLF_self_decode(bits_recovered, S_LENGTH, blocks_no, error_map, KEY, bitstream_ref, reconstruction_map);
            else
                fprintf('WARNING: Decoding method not supported - using self-recovery\n');
                bits_decoder = RLF_self_decode(bits_recovered, S_LENGTH, blocks_no, error_map, KEY, bitstream_ref, reconstruction_map);
            end
            
            % Calculate bitstream errors - take into account only fragments
            % which actually needed to be reconstructed
            if mod(length(bitstream), S_LENGTH) ~= 0
                pad = zeros(1, S_LENGTH - mod(length(bitstream), S_LENGTH));
                bitstream = [bitstream pad];
            end
            for bi = 1:blocks_no
                if reconstruction_map(bi) > 0
                    ERR = ERR + sum(abs(bits_decoder(1+(bi-1)*S_LENGTH:bi*S_LENGTH) - bitstream(1+(bi-1)*S_LENGTH:bi*S_LENGTH)));
                end
            end

        catch e
            fprintf('ERROR %s\n', e.message);
            ERR = length(bitstream);
        end
         
        fprintf('Reference bit-stream errors: %d\n', ERR);
        
        % Reconstruct the Content
        if ERR == 0
            
            imgt_a = double(floor(double(imgt)/dr_factor));
            S_r = blkproc(imgt_a, BLOCK*[1, 1], f);

            data_pointer = 1;
            P = get_reference_alloc(q_bits);
            P = P(indices);
            P(P == 0) = [];
            
            for bx = 1:W/BLOCK
                
                for by = 1:H/BLOCK
                    
                    Ps = sum(sum(P));
                    d = by + (bx-1)*H/BLOCK;

                    if error_map(d) == true
                        data = bits_decoder(data_pointer:data_pointer + Ps - 1);

                        % reconstruct the low quality version of the 
                        % image block content

                        data_index = 1;
                        B = zeros(size(B));

                        for p = 1:length(P)
                            Q = P(p);
                            if p == 1
                                V = binvec2dec(data(data_index:data_index+Q-1));
                            else
                                V = binvec2dec(data(data_index+1:data_index+Q-1));
                            end
                            if p > 1 && data(data_index) == 1
                                V = - V;
                            end
                            B(indices(p)) = V;
                            data_index = data_index + Q;
                        end

                        S_r(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
                    end
                    data_pointer = data_pointer + Ps;
                end
            end

            img_fin = uint8(round(dr_factor*blkproc(S_r, BLOCK*[1, 1], fi)));
            
            % Add artificial noise to restored fragments, if requested
            if RANDOMIZE_LSBS == true
                for bx = 1:W/BLOCK
                    for by = 1:H/BLOCK
                        d = by + (bx-1)*H/BLOCK;
                        if error_map(d) == true
                            B = img_fin(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK);
                            B = B + uint8(randint(BLOCK, BLOCK, 8));
                            img_fin(1+(by-1)*BLOCK:(by)*BLOCK, 1+(bx-1)*BLOCK:(bx)*BLOCK) = B;
                        end
                    end
                end
            else
                % Otherwise, clear the LSBs for the whole image
                img_fin = 8*floor(img_fin/8);
                img_fin = bitset(img_fin,3,1);
            end

            r.rec_psnr = calc_psnr(img, img_fin);        
        end
    end
    
    % Add missing entries to the resulting structure
    r.B = W*H/BLOCK/BLOCK; % number of image blocks
    r.K = uint32(length(bitstream)/S_LENGTH); % number of reference symbols
    r.T = T;  % the number of tampered blocks
    r.E = E;  % error map detection errors
    r.str_errors = ERR; % number of bit-stream errors after decoding
    r.error_map = error_map; % error map
    r.imgt = imgt; % the tampered image
    r.imgt2 = imgt2; % the tampered image with tinted tampering detection
    r.img_fin = img_fin; % final reconstructed image
    r.img_ref = img_low_quality; % the reference image
end