%% An Efficient Method for Content Reconstruction with Self-Embedding
%
% Copyright (C) 2012 Paweł Korus
% All rights reserved.
%
% Version: 2012.07.15
% Contact: Paweł Korus, pkorus@agh.edu.pl
%
% The author is with the Department of Telecommunications of AGH University
% of Science and Technology, Kraków, Poland.
%
% ------------------------------------------------------------------------
%
% This file is a demonstration of a self-embedding scheme based on an 
% erasure communication channel. It accompanies the manuscript: 
%
% P. Korus and A. Dziech, An Efficient Method for Content Reconstruction 
% with Self-Embedding
%
% submitted to IEEE Transactions on Image Processing on 03.01.2012.
%
% This file recreates the most important figures from the paper. Please
% refer to the included README file for instructions.
%
% ------------------------------------------------------------------------
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation version 3.0.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% ----------------------------------------------------------------------
%
% The research leading to these results has received funding from the
% INDECT project funded by European Community's Seventh Framework Pro-
% -gramme (FP7 / 2007-2013) under grant agreement no.[218086].
%

addpath('./include');
% clear; clc;

% The size of the image
image_size = 256;

% The number of iterations during the rec. demand shape estimation
iters = 250;

% The number of iterations for rec. success bound validation
g_iters = 5000;

% Filename for saving the results
filename = 'results/results_20111229_256_250_5000.mat';

%% 1. Reconstruction Success Bounds
%
% The derived bounds on the maximal tampering rate:
%
% * $\gamma_1$ - general erasure channel
% * $\gamma_2$ - optimistic tampering rate bound
% * $\gamma_3$ - estimate of the typical tampering rate 
%                bound for a random tampering pattern
% * $\gamma_3'$ - more accurate typical tamp. rate bound

figure(1); clf;
x = 0.01:0.01:1;
a = 1+1./x;
a2 = 1./x;
y = x./(1+x);
plot(x,1-x, 'k:'); hold on;
plot(x,1-y, 'k--');

p = zeros(size(x));
q = zeros(size(x));
l = zeros(size(x));

for i = x
    % Numerically solve the assumed estimate
    f = @(y) y + i.*y.^(a(x == i)) - i;
    % Numerically solve the assumed estimate
    h = @(y) y + i.*y.^(a2(x == i)) - i;
    % Numerically solve for the more accurate solution
    g = @(y) y + i.* ( ((1/i)-floor(1/i)).* y.^(ceil(1/i)+1) + (1-(1/i)+floor(1/i)).* y.^(ceil(1/i))- 1);
    p(x ==i) = fzero(f, [0 1]);
    q(x ==i) = fzero(g, [0 1]);
    l(x ==i) = fzero(h, [0 1]);
end

plot(x,1-p, 'b-');
plot(x,1-l, 'b:');
plot(x,1-q, 'r-');

xlabel('Code rate \lambda');
ylabel('Tampering rate (1-\gamma)');
legend('$\gamma_1$ bound', '$\gamma_2$ bound', '$\gamma_3$ for $\frac{1}{\lambda} \notin N$', '$\gamma_3$ for $\frac{1}{\lambda} \in N$', '$\gamma_3`$');
h = legend;
set(h, 'interpreter', 'latex');

%% 2. Reconstruction Demand 
%
% The results of the shape parameter $\alpha$ estimation together with the
% derived theoretical results.

% Load a sample Image
img = imread2('testset/lena.png', image_size);

% check if the results are not already available
if ~exist('success_all', 'var') || ~exist('failure_all','var')
    % The considered reconstruction reference sizes, in bits per block
    bs = 8:1:160;

    % For storing the reconstruction results
    success_all = cell(1,length(bs));
    failure_all = cell(1,length(bs));

    % For storing the MLE estimates of the reconstruction demand shape
    shapes = zeros(1,length(bs));

    shi = 1;
    for b = bs

        % For storing the content reconstruction results:
        %  - lambda
        %  - tampering rate gamma
        %  - reconstruction demand rho
        success = zeros(3,iters);
        failure = zeros(3,iters);
        si = 1;
        fi = 1;

        for i = 1:iters

            fprintf('-- Reconstruction demand [b = %d, iteration: %d/%d] --\n', b, i,iters);
            % Select random tampering rate
            g = rand();
            % Perform content reconstruction attempt
            r = reconstruction_test(img, b, g, 'self-recovery', 'random');

            % If there are no errors in the necessary portion of the
            % reconstruction reference, then add the current results
            % to the 'success' array 
            if r.str_errors == 0
               success(:,si) = [b/160; g; r.reconstruction_demand];
               si = si + 1;
            else
                failure(:,fi) = [b/160; g; r.reconstruction_demand];
                fi = fi + 1;
            end
        end

        % Remove last empty portion of the array
        success(:,si:end) = [];
        failure(:,fi:end) = [];

        % Add the current results to a global result container
        success_all{bs == b} = success;
        failure_all{bs == b} = failure;

        % Fit of the reconstruction demand shape parameter
        fx = [success(2,:) failure(2,:)]';
        fy = [success(3,:) failure(3,:)]';
        s = fitoptions('Method', 'NonlinearLeastSquares', 'Startpoint', 2);
        f = fittype('1-(1-x)^a', 'options', s);
        [data_fit gof] = fit(fx,fy,f);

        % Collect the obtained parameter estimate
        shapes(shi) = data_fit.a;
        shi = shi+1;
    end
end

% Plot the obtained shape estimation results along with the factually
% observed samples

figure(1); clf; 
plot(bs./160, shapes,'kd');
hold on;
plot(bs./160, 1./(bs./160), 'k--');
plot(bs./160, 1./(bs./160)+1, 'k-');
plot(bs./160, ceil(1./(bs./160)), 'k:');
xlim([0.2 1]);
ylim([0.75 6]);

ylabel('Reconstruction demand shape \alpha');
xlabel('Code rate \lambda');
legend('$\hat{\alpha}$ estimates', '$\frac{1}{\lambda}$', '$1+\frac{1}{\lambda}$', '$\lceil 1+\frac{1}{\lambda} \rceil$');
h = legend;
set(h, 'interpreter', 'latex');


%% 3. Reconstruction Demand for Selected Code Rates
%
% Experimental vs. theoretical dependency between the reconstruction demand
% and the tampering rate for selected reference payloads $\lambda$. The 
% corresponding $\gamma_1$ and $\gamma_3$ are plotted using dotted lines. 
% Success cases are marked with circles, failures are marked with crosses.

% Include the following reconstruction reference rates
bsp = [160, 80, 40, 120];

figure(1); clf;

for bi = 1:length(bsp)
    
    charts_y = ceil(sqrt(length(bsp)));
    charts_x = ceil(length(bsp)/charts_y);
    subplot(charts_y, charts_x, bi);
%     figure(bi);
    bsi = bsp(bi);
    
    % Fetch reconstruction results for the selected case
    success = success_all{bs == bsi};
    failure = failure_all{bs == bsi};

    % Plot reconstruction demand vs. tampering rate
    % Differentiate success and failure cases
    plot(success(2,:), success(3,:), 'o', failure(2,:), failure(3,:), 'x');
    hold on;
    
    % Add a previously fitted average reconstruction demand 
    plot(x,1-(1-x).^shapes(bs == bsi), '-k');

    xlabel('Tampering rate (1-\gamma)');
    ylabel('Reconstruction demand \rho');
    
    % Add the analytically derived reconstruction success bound
    lambda = success(1,1);
    
    % The gamma_1 bound
    plot((1-lambda)*[1 1], [0 1], 'k:');
    
    % The gamma_2 bound
    plot(1-lambda/(1+lambda)*[1 1], [0 1], 'k--');
    
    % The gamma_3 bound - the assumed estimate
    alpha = 1+1/lambda;
    if round(1/lambda) == 1/lambda
        alpha = 1/lambda;
    end
    
    title(sprintf('B = %d bits ($\\lambda$ = %.2f), $\\alpha$: %.2f, $\\hat{\\alpha}$: %.2f', bsi, bsi/160, alpha, shapes(bs == bsi)),'interpreter', 'latex');
    
    f = @(y) y + lambda.*y.^(alpha) - lambda;
    g3 = 1 - fzero(f, [0 1]);
    plot(g3*[1 1], [0 1], 'r-');
    
    % The gamma_3 bound - the fitted estimate
    f = @(y) y + lambda.*y.^(shapes(bs == bsi)) - lambda;
    g3 = 1 - fzero(f, [0 1]);
    plot(g3*[1 1], [0 1], 'r:');
end

%% 4. Experimental Evaluation of the Rec. Bound - Gamma 1
%
% The goal of this experiment is to verify the analytical reconstruction
% performance bound $\gamma_1$. A general erasure channel model is used for
% decoding the reference payload. Each sample in the plot represents a
% single reconstruction attempt. The applicable theoretical bound between 
% the successes and failures is shown as a solid line. 

% Load the test image, make sure it's in grayscale, resize
img = imread2('testset/lena.png', image_size);

% check if the results are not already available
if ~exist('g1_success', 'var') || ~exist('g1_failure','var')
    
    % Initial setup
    g1_success = zeros(3,iters);
    g1_failure = zeros(3,iters);

    si = 1;
    fi = 1;

    for i = 1:g_iters

        fprintf('-- Gamma 1 Bound [iteration: %d / %d]--\n', i, g_iters);

        % Select random reconstruction reference quality
        b = max(5,round(160 * rand()));

        % Select random tampering rate
        g = rand();

        % Attempt to reconstruct with the selected settings
        r = reconstruction_test(img, b, g, 'general-erasure', 'random');

        % Update the appropriate array with the obtained result
        if r.str_errors == 0
            g1_success(:,si) = [b/160; g; r.reconstruction_demand];
            si = si + 1;
        else
            g1_failure(:,fi) = [b/160; g; r.reconstruction_demand];
            fi = fi + 1;
        end
    end

    % Remove unused elements of the arrays
    g1_success(:,si:end) = [];
    g1_failure(:,fi:end) = [];
end

% Plot the results
figure(1); clf;
plot(g1_success(1,:), g1_success(2,:), 'o', g1_failure(1,:), g1_failure(2,:), 'x'); 
hold on;
plot(x,1-x, 'k-');
plot(x,1-y, 'k--');
plot(x,1-p, 'k-.');

xlabel('Code rate \lambda');
ylabel('Tampering rate (1-\gamma)');
legend('success', 'failure', '\gamma_1', '\gamma_2', '\gamma_3');


%% 5. Experimental Evaluation of the Rec. Bound - Gamma 2
%
% The goal of this experiment is to verify the analytical reconstruction
% performance bound $\gamma_2$ (continuous tampering pattern). 
% The proposed self-recovery model is used for decoding the reference 
% payload. Each sample in the plot represents a single reconstruction 
% attempt. The applicable theoretical bound between the successes and 
% failures is shown as a solid line. 

% Load the test image, make sure it's in grayscale, resize
img = imread2('testset/lena.png', image_size);

% check if the results are not already available
if ~exist('g2_success', 'var') || ~exist('g2_failure','var')
    % Initial setup
    g2_success = zeros(3,iters);
    g2_failure = zeros(3,iters);

    si = 1;
    fi = 1;

    for i = 1:g_iters

        fprintf('-- Gamma 2 Bound [iteration: %d / %d]--\n', i, g_iters);

        % Select random reconstruction reference quality
        b = max(5,round(160 * rand()));

        % Select random tampering rate
        g = rand();

        % Attempt to reconstruct with the selected settings
        r = reconstruction_test(img, b, g, 'self-recovery', 'continuous');

        % Update the appropriate array with the obtained result
        if r.str_errors == 0
            g2_success(:,si) = [b/160; g; r.reconstruction_demand];
            si = si + 1;
        else
            g2_failure(:,fi) = [b/160; g; r.reconstruction_demand];
            fi = fi + 1;
        end
    end

    % Remove unused elements of the arrays
    g2_success(:,si:end) = [];
    g2_failure(:,fi:end) = [];
end

% Plot the results
figure(1); clf;
plot(g2_success(1,:), g2_success(2,:), 'o', g2_failure(1,:), g2_failure(2,:), 'x'); 
hold on;
plot(x,1-x, 'k--');
plot(x,1-y, 'k-');
plot(x,1-p, 'k-.');

xlabel('Code rate \lambda');
ylabel('Tampering rate (1-\gamma)');
legend('success', 'failure', '\gamma_1', '\gamma_2', '\gamma_3');


%% 6. Experimental Evaluation of the Rec. Bound - Gamma 3
%
% The goal of this experiment is to verify the analytical reconstruction
% performance bound $\gamma_3$ (random tampering pattern). 
% The proposed self-recovery model is used for decoding the reference 
% payload. Each sample in the plot represents a single reconstruction 
% attempt. The applicable theoretical bound between the successes and 
% failures is shown as a solid line. 

% Load the test image, make sure it's in grayscale, resize
img = imread2('testset/lena.png', image_size);

% check if the results are not already available
if ~exist('g3_success', 'var') || ~exist('g3_failure','var')
    % Initial setup
    g3_success = zeros(3,iters);
    g3_failure = zeros(3,iters);

    si = 1;
    fi = 1;

    for i = 1:g_iters

        fprintf('-- Gamma 3 Bound [iteration: %d / %d]--\n', i, g_iters);

        % Select random reconstruction reference quality
        b = max(5,round(160 * rand()));

        % Select random tampering rate
        g = rand();

        % Attempt to reconstruct with the selected settings
        r = reconstruction_test(img, b, g, 'self-recovery', 'random');

        % Update the appropriate array with the obtained result
        if r.str_errors == 0
            g3_success(:,si) = [b/160; g; r.reconstruction_demand];
            si = si + 1;
        else
            g3_failure(:,fi) = [b/160; g; r.reconstruction_demand];
            fi = fi + 1;
        end
    end

    % Remove unused elements of the arrays
    g3_success(:,si:end) = [];
    g3_failure(:,fi:end) = [];
end

% Plot the results
figure(1); clf;
plot(g3_success(1,:), g3_success(2,:), 'o', g3_failure(1,:), g3_failure(2,:), 'x'); 
hold on;
plot(x,1-x, 'k--');
plot(x,1-y, 'k-.');
plot(x,1-p, 'k-');

xlabel('Code rate \lambda');
ylabel('Tampering rate (1-\gamma)');
legend('success', 'failure', '\gamma_1', '\gamma_2', '\gamma_3');

%% Save the results
fprintf(' INFO Saving results to "%s"\n', filename)
save(filename);